import { select, selectAll, scaleOrdinal } from 'd3';

// Setting up the main SVG container properties
const svg = select('svg');
const width = +svg.attr('width');
const height = +svg.attr('height');
const centre_x = width / 2; // Central X position for positioning nodes
const centre_y = height / 2; // Central Y position for positioning nodes
const colorScale = ['#FF338A']; // Color scale for nodes, currently only one color

// Configuring the SVG's aspect ratio and view box for responsive design
svg
  .attr('preserveAspectRatio', 'xMidYMid meet')
  .attr('viewBox', [0, 0, width * 2, height * 2]);

// Defining SVG definitions for markers, like arrowheads
const defs = svg.append('defs');
defs
  .selectAll('marker')
  .data(['node']) // Data is just a single marker named 'node'
  .enter()
  .append('marker') // Appending a marker for nodes
  .attr('id', (d) => d) // Setting ID based on data, here it's 'node'
  .attr('viewBox', '0 -2.5 5 5') // Viewbox for the marker
  .attr('refX', 25) // Reference position for the marker on the X-axis
  .attr('refY', 0) // Reference position for the marker on the Y-axis
  .attr('markerWidth', 5) // Marker width
  .attr('markerHeight', 9) // Marker height
  .attr('orient', 'auto') // Orientation is automatic
  .append('path') // Appending a path element to the marker
  .attr('d', 'M0,-2.5L5,0L0,2.5') // Path command for an arrow shape
  .attr('fill', 'red'); // Fill color for the marker

// Loading data from a JSON file
d3.json('data.json').then((data) => {
  // Processing links from the loaded data
  const links = data.links.map((d) => Object.create(d));
  links.forEach((d, i) => {
    d.srcType = data.links[i].srcType;
    d.tgtType = data.links[i].tgtType;
    d.relationship = data.links[i].relationship;
    d.value = data.links[i].value;
  });

  // Processing nodes from the loaded data
  const nodes = data.nodes.map((d) => Object.create(d));
  nodes.forEach((d, i) => {
    d.group = data.nodes[i].group;
  });

  // Setting up the force simulation for the nodes and links
  const simulation = d3
    .forceSimulation(nodes) // Initiating force simulation with nodes data
    .force(
      'link',
      d3
        .forceLink(links) // Applying link force with links data
        .id((d) => d.id) // Identifying links by node id
        .distance(120), // Setting distance between linked nodes
    )
    .force('charge', d3.forceManyBody().strength(-500)) // Repulsive force between nodes
    .force('collision', d3.forceCollide().radius(105)) // Collision force to prevent node overlap
    .force('center', d3.forceCenter(width, height)); // Centering force to position nodes around the center

  // Drag behavior for nodes
  const drag = (simulation) => {
    function dragstarted(event) {
      if (!event.active)
        simulation.alphaTarget(0.5).restart(); // Restarting simulation with a higher alpha target on drag start
      event.subject.fx = event.subject.x; // Fixing node position on X-axis
      event.subject.fy = event.subject.y; // Fixing node position on Y-axis
    }
    function dragged(event) {
      event.subject.fx = event.x; // Updating fixed position on X-axis while dragging
      event.subject.fy = event.y; // Updating fixed position on Y-axis while dragging
    }
    function dragended(event, d) {
      if (!event.active) simulation.alphaTarget(0); // Resetting alpha target on drag end
      event.subject.fx = null; // Unfixing node position on X-axis
      event.subject.fy = null; // Unfixing node position on Y-axis
    }
    return d3
      .drag()
      .on('start', dragstarted)
      .on('drag', dragged)
      .on('end', dragended);
  };

  // Creating link elements in the SVG
  const link = svg
    .append('g')
    .selectAll('line')
    .data(links)
    .join('line')
    .attr('marker-end', 'url(#node)'); // Attaching markers to the end of links

  // Creating node elements in the SVG
  const node = svg
    .append('g')
    .attr('class', 'node')
    .selectAll('g')
    .data(nodes)
    .enter()
    .append('g')
    .attr(
      'transform',
      (d) =>
        'translate(' + centre_x + ', ' + centre_y + ')', // Initial positioning at the center
    );

  // Appending circles to represent nodes
  const circles = node
    .append('circle')
    .attr('r', 75) // Radius of nodes
    .attr('fill', (d) => colorScale[0]) // Fill color from the color scale
    .on('mouseover', focusNode) // Event listener for mouseover
    .on('mouseout', blurNode) // Event listener for mouseout
    .call(drag(simulation)); // Applying drag behavior

  // Adding text labels to nodes
  const nodeLbl = node
    .append('text')
    .text((d) => d.id) // Setting text to node id
    .attr('dy', '0.3em'); // Vertical offset for text

  // Updating positions on simulation 'tick'
  simulation.on('tick', () => {
    link
      .attr('x1', (d) => d.source.x) // Setting x1 position of line to source node x position
      .attr('y1', (d) => d.source.y) // Setting y1 position of line to source node y position
      .attr('x2', (d) => d.target.x) // Setting x2 position of line to target node x position
      .attr('y2', (d) => d.target.y); // Setting y2 position of line to target node y position
    node.attr(
      'transform',
      (d) => 'translate(' + d.x + ', ' + d.y + ')', // Updating node position
    );
  });

  // Function to highlight node on mouseover
  function focusNode() {
    d3.select(this).classed('nodeHover', true);
  }

  // Function to remove highlight from node on mouseout
  function blurNode() {
    d3.select(this).classed('nodeHover', false);
  }
});
